#Importing modules ####
library(Seurat)
library(DUBStepR)
library(ggplot2)
library(profvis)
library(profmem)
library(scran)
library(gridExtra)
library(grid)
library(scales)
library(ComplexHeatmap)
library(circlize)
library(Matrix)
library(qlcMatrix)
library(reshape2)
library(ggplotify)
library(ggrepel)
library(dplyr)
library(gridGraphics)
library(umap)
library(Rcpp)
library(Hmisc)
library(SingleCellExperiment)
library(scater)
library(lubridate)
library(cluster)
library(plotly)
library(org.Hs.eg.db)
library(org.Mm.eg.db)
library(scRNAseq)
library(M3Drop)
library(rhdf5)

#Sourcing Helper Functions #######

source("parallelBenchmarkingFunctions.R")
source("trendVar.R")
source("UseM3D.R")
source("UseHEG.R")
source("UseHVG2.R")
source("GiniClust_Fitting.R")

# Preprocessing ########
message(" ")
message(" ")
message("Loading datasets")

preprocess_dataset_fn <- function(obj) {
  sce <- obj
  sce <- sce[, colSums(counts(sce)) > 0]
  sce <- scuttle::logNormCounts(sce)
  rownames(sce) <- make.unique(rownames(sce))
  return(list("Object" = sce, "Matrix" = sce@assays@data$logcounts))
}

mouse_cell_10k <- preprocess_dataset_fn(obj = readRDS(file = "mouse_cell_atlas_symbols_downsampled_10k.RDS"))

mouse_cell_100k <- preprocess_dataset_fn(obj = readRDS(file = "mouse_cell_atlas_symbols_downsampled_100k.RDS"))

mouse_cell_1M <- preprocess_dataset_fn(obj = readRDS(file = "mouse_cell_atlas_symbols_original_1M.RDS"))


dataset_name <- "ADD DATASET NAME"
out_dir <- "OUTPUT PATH"
MC_CORES <- "NUMBER OF CORES AVAILABELE"
  
# Pass in processed dataset here. 10k dataset is given here for reference
testMatrix <- mouse_cell_10k$Matrix
testMatrixObj <- mouse_cell_10k$Object
  
message(" ")
message(" ")
message(paste("Output files will be saved in", out_dir))
message("Datasets Loaded")

#Benchmarking function ####
benchmark <- function(function_name, func, source_files = F, source_directory = "", data, output_directory, dataset_name){
  if(function_name == "DUBStepR"){
      message(" ")
      message(" ")
      message(paste("Benchmarking", function_name))
      message(" ")
      message(" ")
      Rprof(filename = paste0(output_directory, function_name,".prof"), memory.profiling = T)
      st <- system.time({
      out <-
        func(
          input.data = data,
          optimise.features = F
        )
      })
      Rprof(NULL)
      message(" ")
      message(" ")
      message(paste0("Saving output file ", function_name,".prof"))
      message("Benchmarking Complete")
      return(list("genes" = out$corr.info$feature.genes, "time" = st))
  }
  else {
    message(" ")
    message(" ")
    message(paste("Benchmarking", function_name))
    message(" ")
    message(" ")
    Rprof(filename = paste0(output_directory, function_name,".prof"), memory.profiling = T)
      out <- tryCatch(
        {
          func(data = data, data_name = dataset_name)
        }, 
        error = function(cond){
          message("Encountered error, returning empty list")
          message(" ")
          message(cond)
          return(list("genes" = "NA", "time" = "NA"))
        }
      )
    Rprof(NULL)
    message(" ")
    message(" ")
    message(paste0("Saving output file ", function_name,".prof..."))
    message("Benchmarking Complete")
    return(list("genes" = out$genes, "time" = out$time))
  }
}


#Running Benchmarking function with each algorithm ######
bench_out_dubstepr <- benchmark(function_name = "DUBStepR", func = DUBStepR, data = testMatrix, dataset_name = dataset_name, output_directory = out_dir)
bench_out_hvgdisp <- benchmark(function_name = "HVGDisp", func = hvg_disp_fn, data = testMatrix, dataset_name = dataset_name, output_directory = out_dir)
bench_out_hvgvst <- benchmark(function_name = "HVGVST", func = hvg_vst_fn, data = testMatrix, dataset_name = dataset_name, output_directory = out_dir)
bench_out_trendvar <- benchmark(function_name = "trendVar", func = trendVar_fn, data = testMatrixObj, dataset_name = dataset_name, output_directory = out_dir)
bench_out_m3dropdanb <- benchmark(function_name = "M3DropDANB", func = m3drop_fn, data = testMatrixObj@assays@data$counts, dataset_name = dataset_name, output_directory = out_dir)
bench_out_gini <- benchmark(function_name = "Giniclust", func = gini_fn, data = testMatrix, dataset_name = dataset_name, output_directory = out_dir)
bench_out_hlg <- benchmark(function_name = "HLG", func = pca_fn, data = testMatrix, dataset_name = dataset_name, output_directory = out_dir)

bench_out_list <- list(
                         "DUBStepR" = bench_out_dubstepr,
                         "HVGDisp" =  bench_out_hvgdisp,
                         "HVGVST" = bench_out_hvgvst,
                         "trendVar" = bench_out_trendvar,
                         "M3DropDANB" = bench_out_m3dropdanb,
                         "Giniclust" = bench_out_gini,
                         "HLG"= bench_out_hlg)

saveRDS(bench_out_list, file = paste0(out_dir,"benchmark_output.rds"))

plot_graphs <- function(directory) {
  mem_files <- list.files(directory, pattern = ".prof")
  benchmark_output <- readRDS(paste0(directory, "benchmark_output.rds"))
  
  funcs <- c()
  times <- c()
  mem_usage <- c()
  
  for (file in mem_files) {
    func <- gsub(x = file,
                 pattern = ".prof",
                 replacement = "")
    path <- paste(directory, file, sep = "")
    profiling_out <- summaryRprof(path, memory = "both")
    funcs <- append(funcs, func)
    if (benchmark_output[[func]]$time != "NA"){
    times <-
      append(times, benchmark_output[[func]]$time["elapsed"])
    mem_usage <-
      append(mem_usage, profiling_out[["by.total"]][["mem.total"]][2])
    } else {
      times <-
        append(times, 0)
      mem_usage <-
        append(mem_usage, 0)
    }
    
    
  }
  times <- times / 60
  mem_usage <- mem_usage / 8000
  func_times <- data.frame("Function" = funcs, "Time" = times)
  time_plot <-
    ggplot(func_times, aes(x = Function, y = Time)) + geom_bar(stat = "identity") + coord_flip() + scale_y_continuous(labels = scales::comma) + labs(x = "Functions", y = "Running Time (minutes)")
  
  memory_allocations <- data.frame("Function" = funcs, "MemoryUsage" = mem_usage)
  
  memory_plot <-
    ggplot(memory_allocations, aes(x = Function, y = MemoryUsage)) + geom_bar(stat = "identity") + coord_flip() + scale_y_continuous(labels = scales::comma) + labs(x = "Functions", y = "Memory Usage (GB)") +
    theme(
      axis.text.y = element_blank(),
      axis.ticks.y = element_blank(),
      axis.title.y = element_blank(),
      plot.background = element_blank()
    )
  
  return(list("Run_Time_plot" = time_plot, "Run_Time" = func_times, "Memory_Usage_plot" = memory_plot, "Memory_Usage" = memory_allocations))
}




p <- plot_graphs(out_dir)

p$Run_Time
p$Memory_Usage


grid.arrange(p$Run_Time_plot, p$Memory_Usage_plot, ncol=2, top=textGrob(dataset_name))

